from typing import Any, Dict
import yaml
import os
from testcompose.models.config import ITestConfig


class TestConfigParser:
    @classmethod
    def parse_config(cls, file_name: str) -> ITestConfig:
        """parses and verifies test yaml config file

        Args:
            file_name (str): absolutel path of the config file

        Raises:
            FileNotFoundError: when config file not present
            AttributeError: when config file is empty

        Returns:
            ITestConfig: A ITestConfig object with all named services in the config
        """
        if not os.path.exists(file_name):
            raise FileNotFoundError(f"Config file {file_name} does not exist!!")

        content: Dict[str, Any] = dict()
        with open(file_name, 'r') as fh:
            content = yaml.safe_load(fh)

        if not content:
            raise AttributeError(f"Config content can not be empty")

        test_services: ITestConfig = ITestConfig(**content)
        return test_services



import time
from uuid import uuid4
from typing import Any, Dict, List, Optional, Tuple
from docker.client import DockerClient
from testcompose.containers.generic_container import GenericContainer
from testcompose.containers.container_network import ContainerNetwork
from .models.container import RunningContainer, RunningContainerKeys, RunningContainerEnvPrefixes
from .containers.utils import ContainerUtils
from .models.config import RankedServices, ITestConfigServices
from .models.client import ClientFromEnv, ClientFromUrl, Login
from .models.container import ContainerParam
from .client.client_from_env import EnvClient
from .client.client_from_url import UrlClient
from .log_setup import stream_logger


logger = stream_logger(__name__)


class RunContainers:
    """Run test containers

    Args:
        services (RankedServices): Ranked container services to be ran
        wait_time_between_container_start (float, optional): Default time to wait before spawning the next container.
                                                             This is aimed at preventing starting containers too earlier
                                                             before the container(s) they depend on has not fully started.
                                                             Especially when they do not have a log predicate to search for
                                                             or an http endpoint to check agains.
                                                             Defaults to 20.0.
        url_param (ClientFromUrl, optional): Url parameters for connecting to the docker daemon.
                                             Defaults to ClientFromUrl().
        env_param (ClientFromEnv, optional): Environment variables parameter required to connect to
                                             the docker daemon.
                                             Defaults to ClientFromEnv().
        registry_login_param ([type], optional): Parameter providing docker registry login details.
                                             Defaults to Login().
    """

    def __init__(
        self,
        services: RankedServices,
        wait_time_between_container_start: float = 20.0,
        url_param: ClientFromUrl = ClientFromUrl(),
        env_param: ClientFromEnv = ClientFromEnv(),
        registry_login_param=Login(),
    ) -> None:
        self._ranked_services: RankedServices = services
        self._running_containers: RunningContainer = RunningContainer()
        self._dependency_mapping: Dict[str, int] = dict()
        self._running_containers_extra_envs: Dict[str, Dict[str, Any]] = dict()
        self._wait_time_between_container_start: float = wait_time_between_container_start
        self._registry_login_param: Login = registry_login_param
        self.dclient = self.get_docker_client(env_param, url_param)
        self._container_network: Optional[ContainerNetwork] = None

    def get_docker_client(self, env_param: ClientFromEnv, url_param: ClientFromUrl) -> DockerClient:
        """Docker client.
        `Note`: only one of the parameters need be provided. If both the
        env and url parameters are provided, the url parameter takes precedence.

        Args:
            env_param (ClientFromEnv): Environment variables parameter required to connect to
                                       the docker daemon.
            url_param (ClientFromUrl): Url parameters for connecting to the docker daemon

        Raises:
            Exception: When connection to docker daemon fails

        Returns:
            DockerClient: docker client
        """
        try:
            client = EnvClient(env_param) if not url_param.docker_host else UrlClient(url_param)
            client.initialise_client()
            assert client.docker_client.ping()
            return client.docker_client
        except Exception as exc:
            logger.error("%s", exc)
            raise Exception

    @property
    def dclient(self) -> DockerClient:
        """docker client

        Returns:
            DockerClient: docker client
        """
        return self._docker_client

    @dclient.setter
    def dclient(self, client: DockerClient) -> None:
        self._docker_client = client

    @property
    def running_containers(self) -> RunningContainer:
        """Running container objects

        Returns:
            RunningContainer: Running containers
        """
        return self._running_containers

    @running_containers.setter
    def running_containers(self, containers: RunningContainer):
        self._running_containers = containers

    @property
    def unique_container_label(self) -> str:
        """Unique test string label.

        Returns:
            str: unique label
        """
        return self._unique_container_label

    @unique_container_label.setter
    def unique_container_label(self, label: str):
        self._unique_container_label = label

    def __enter__(self) -> RunningContainer:
        return self.run()

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.stop_running_containers()
        if exc_tb and exc_type:
            logger.info("%s[%s]: %s", exc_type, exc_value, exc_tb)

    def get_extra_dependency_envs(
        self, service: ITestConfigServices, network_name: Optional[str] = None
    ) -> Tuple[Dict[str, Dict[str, Any]], List[str]]:
        """Additional environment variables derived for easy manipulation of the
        running containers. These are variables extracted from already running
        containers to which the current container depends.

        Args:
            service (ITestConfigServices): Service to assigne special variable
            network_name (Optional[str], optional): Container network name.
                                                    Defaults to None.

        Raises:
            AttributeError: When a cyclic container dependency is detected.

        Returns:
            Tuple[Dict[str, Dict[str, Any]], List[str]]: A tuple of `extra_env` and `modified_exposed_ports`
        """
        extra_env: Dict[str, Dict[str, Any]] = dict()
        extra_env_dict_keys: Dict[str, Any] = dict()
        if service.name in self._dependency_mapping:
            raise AttributeError("A container can not have a cyclic dependency on itself")

        for depends in service.depends_on:
            if depends in self._dependency_mapping:
                container: GenericContainer = self.running_containers.containers[depends]
                extra_env_dict_keys.update(
                    {
                        f"{RunningContainerEnvPrefixes.INTERNAL_HOST}_{depends.upper()}": container.get_container_ip(
                            network_name=network_name
                        ),
                        f"{RunningContainerEnvPrefixes.EXTERNAL_HOST}_{depends.upper()}": container.get_container_host_ip(),
                        f"{RunningContainerEnvPrefixes.MAPPED_PORTS}_{depends.upper()}": container.get_mapped_container_ports(
                            ports=self._ranked_services.services[
                                (self._dependency_mapping[depends], depends)
                            ].exposed_ports
                        ),
                    }
                )
        extra_env_dict_keys.update({**service.environment})
        env_config, modified_exposed_ports = ContainerUtils.replace_container_config_placeholders(
            env_config=extra_env_dict_keys,
            dependency_mapping=self._dependency_mapping,
            running_containers=self.running_containers,
            unique_container_label=self.unique_container_label,
            service_name=service.name,
            exposed_ports=service.exposed_ports,
        )
        extra_env.update({service.name: env_config})
        logger.info("Found extra env variables: %s", extra_env)
        return extra_env, modified_exposed_ports

    def get_container_envs(self, container: GenericContainer, exposed_ports: List[str]) -> Dict[str, Any]:
        """These are environment variables included in all running
        containers even when their configuration does not contain
        any environment variable. These are required to allow for
        some certain level of interactions with the container at runtime.

        Args:
            container (GenericContainer): Container object
            exposed_ports (List[str]): Container exposed port

        Returns:
            Dict[str, Any]: extra container envirnment variables
        """
        return {
            RunningContainerEnvPrefixes.EXTERNAL_HOST: container.get_container_host_ip(),
            RunningContainerEnvPrefixes.MAPPED_PORTS: container.get_mapped_container_ports(
                ports=exposed_ports
            ),
        }

    def run(self) -> RunningContainer:
        """Run all Generic Test Containers

        Raises:
            AttributeError: When current container being stated had a record it was already running.
            TimeoutError: When a predicate could not be found in the log over a specified period of time
            RuntimeError: When an http endpoint did not return the expected specified status code.
            Exception: When an Unknown error occured during the process of starting a container.

        Returns:
            RunningContainer: running containers
        """
        self.unique_container_label = uuid4().hex
        auto_create_network = (
            self._ranked_services.network.auto_create
            if not self._ranked_services.network.use_random_network
            else self._ranked_services.network.use_random_network
        )
        network_name = (
            self._ranked_services.network.name
            if not self._ranked_services.network.use_random_network
            else f"{self.unique_container_label}_network"
        )
        self._container_network = ContainerNetwork(
            docker_client=self.dclient, network_name=network_name, auto_create_network=auto_create_network
        )
        running_containers: Dict[str, GenericContainer] = dict()
        try:
            ranks: List[Tuple[int, str]] = sorted(self._ranked_services.services)
            for rank, name in ranks:
                service = self._ranked_services.services[(rank, name)]
                extra_envs, _modified_container_exposed_ports = self.get_extra_dependency_envs(
                    service, self._container_network.network_name
                )
                generic_container: GenericContainer = GenericContainer(
                    docker_client=self.dclient,
                    container_param=ContainerParam(
                        image=service.image,
                        exposed_ports=_modified_container_exposed_ports,
                        command=service.command,
                        entry_point=service.entrypoint,
                        environment_variables=extra_envs.get(service.name),
                        volumes=service.volumes,
                        auto_remove=service.auto_remove,
                        remove_container=service.remove_container,
                        registry_login_param=self._registry_login_param,
                    ),
                    labels=[f"{self.unique_container_label}_{service.name}"],
                    hostname=f"{self.unique_container_label}_{service.name}",
                    network=self._container_network.network_name,
                )
                generic_container.start()
                running_containers.update({name: generic_container})
                con_extra_env: Dict[str, Any] = self.get_container_envs(
                    generic_container, service.exposed_ports
                )
                generic_container.env = {**con_extra_env, **extra_envs.get(service.name, {})}
                self._dependency_mapping.update({name: rank})

                self._running_containers_extra_envs[service.name] = generic_container.env
                self.running_containers = RunningContainer(
                    **{
                        RunningContainerKeys.CONTAINERS: running_containers,
                        RunningContainerKeys.EXTRA_ENVS: self._running_containers_extra_envs,
                        RunningContainerKeys.PRECEDENCE_MAP: self._dependency_mapping,
                    }
                )
                if (rank, service.name) in running_containers:
                    logger.info("Duplicate name found in running containers : %s", service.name)
                    raise AttributeError

                if service.log_wait_parameters:
                    if service.log_wait_parameters.log_line_regex:
                        found: bool = generic_container.search_container_logs(
                            search_string=service.log_wait_parameters.log_line_regex,
                            timeout=float(service.log_wait_parameters.wait_timeout or 60),
                            interval=int(service.log_wait_parameters.poll_interval or 1),
                        )
                        if not found:
                            raise TimeoutError(
                                "container %s did not emit logs satisfying predicate in %.3f seconds"
                                % (service.name, float(service.log_wait_parameters.wait_timeout or 60))
                            )

                if service.http_wait_parameters:
                    if (
                        service.http_wait_parameters.http_port
                        and service.http_wait_parameters.response_status_code
                    ):
                        response = generic_container.with_wait_for_http(
                            http_port=str(service.http_wait_parameters.http_port),
                            status_code=service.http_wait_parameters.response_status_code,
                            end_point=service.http_wait_parameters.end_point,
                        )
                        if not response:
                            raise RuntimeError(
                                f"Http check on port {service.http_wait_parameters.http_port} failed"
                            )

                time.sleep(self._wait_time_between_container_start)

        except Exception as exc:
            self.stop_running_containers()
            logger.info("%s", exc)
            raise Exception
        logger.info(
            "The following containers were started: %s",
            [x.name for x in self._ranked_services.services.values()],
        )
        return self.running_containers

    def stop_running_containers(self):
        """Stop all running containers"""
        _ranks: List[Tuple[int, str]] = sorted(
            [(y, x) for x, y in self._dependency_mapping.items()], reverse=True
        )
        try:
            for _rank, _name in _ranks:
                container: GenericContainer = self.running_containers.containers[_name]
                container.stop()
                logger.info(
                    "Successfully stopped container: %s(%s): %s",
                    _name,
                    str(_rank),
                    container.get_container_id(),
                )
                time.sleep(5)
        finally:
            if (
                any(
                    [
                        self._ranked_services.network.auto_create,
                        self._ranked_services.network.use_random_network,
                    ]
                )
                and self._container_network
            ):
                self._container_network.remove_network()
            logger.info("Run the ryuk container here!!!")



import re
import socket
from copy import copy, deepcopy
from typing import Any, Dict, List, Optional, Tuple
from testcompose.models.container import RunningContainer


class ContainerUtils:
    @staticmethod
    def replace_container_config_placeholders(
        env_config: Dict[str, Any],
        dependency_mapping: Dict[str, int],
        running_containers: RunningContainer,
        unique_container_label: str,
        service_name: str,
        exposed_ports: List[str],
    ) -> Tuple[Dict[str, Any], List[str]]:
        """Replace placeholders in the service. Placeholders are
        usually of the form *${container_name.containerenv_variable}*.

        Args:
            env_config (Dict[str, Any]): Dict of config environment variables
            dependency_mapping (Dict[str, int]): Dict showing service dependencies
            running_containers (RunningContainer): Running container object
            unique_container_label (str): uuid string for uniquely labelling a given test run
            service_name (str): service name as specified in the config
            exposed_ports (List[str]): container exposed ports

        Raises:
            ValueError: when a placeholder variable is not of the form service_name.variable_name
            AttributeError: when container service name is not present in `dependency_mapping`
            AttributeError: If a test is attempting to replace the exposed port on the host ar runtimr, this
                            should be of the form hostport_1234. Where 1234 is the container exposed port.
                            The container exposed port must be present in the list of `exposed_ports`
            AttributeError: When substituted variable is not one of [`self`, `hostname`, `hostport_1234`].
                            Where 1234 is the container exposed port.

        Returns:
            Tuple[Dict[str, Any], List[str]]: A tuple of `env_config` and `exposed_ports`
        """
        pattern = "\\$\\{([^}]*)}"
        _env_config: Dict[str, Any] = copy(env_config)
        _exposed_ports = deepcopy(exposed_ports)
        cmpl = re.compile(pattern=pattern).findall
        for k, v in env_config.items():
            if isinstance(v, str):
                replaced_varoable = v
                for occurence in cmpl(v):
                    if len(str(occurence).split(".")) != 2:
                        raise ValueError
                    container_name, variable_name = str(occurence).split(".")
                    if container_name.lower() not in dependency_mapping and container_name.lower() != "self":
                        raise AttributeError
                    _value = None
                    if container_name.lower() != "self":
                        if variable_name.lower().startswith("docker_python"):
                            _value = env_config[f"{variable_name.upper()}_{container_name.upper()}"]
                        else:
                            _value = running_containers.containers[f"{container_name.lower()}"].env[
                                f"{variable_name.upper()}"
                            ]
                    else:
                        if variable_name.lower() == "hostname":
                            _value = f"{unique_container_label}_{service_name}"
                        elif variable_name.lower().startswith("hostport_"):
                            container_port = re.sub("hostport_", "", variable_name)
                            host_port = ContainerUtils._get_assigned_free_port(_exposed_ports, container_port)
                            if not host_port:
                                host_port = ContainerUtils.get_free_host_port()
                                if container_port and container_port not in _exposed_ports:
                                    raise AttributeError(
                                        f"self.hostport_{container_port} must be a valid supplied exposed_ports value!"
                                    )
                                _exposed_ports.remove(container_port)
                                _exposed_ports.append(f"{host_port}:{container_port}")
                            _value = str(host_port)
                        else:
                            raise AttributeError(
                                "Container self reference only support hostname and hostport. i.e self.hostname|self.hostport_1234"
                            )
                    replaced_varoable = replaced_varoable.replace(f"${{{occurence}}}", str(_value))
                _env_config[k] = replaced_varoable
        return _env_config, _exposed_ports

    @staticmethod
    def get_free_host_port() -> str:
        """Get a free random port number from the container host

        Returns:
            str: port number
        """
        s = socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        s.settimeout(2)
        s.bind(("", 0))
        addr, port = s.getsockname()
        s.close()
        return port

    @staticmethod
    def _get_assigned_free_port(modified_exposed_ports: List[str], container_port: str) -> Optional[str]:
        """Assigne host:container port system from an obtained
        free random port number from the container host

        Args:
            modified_exposed_ports (List[str]): A placeholder host port beginning with `hostport_`
            container_port (str): Exposed container port

        Returns:
            Optional[str]: host port number
        """
        host_port: Optional[str] = None
        count = set([x for x in modified_exposed_ports if x.endswith(f":{container_port}")])
        if count:
            host_port = str(list(count)[0]).split(":")[0]

        return host_port



from docker import DockerClient
from abc import ABC, abstractmethod
from docker.constants import DEFAULT_TIMEOUT_SECONDS, DEFAULT_MAX_POOL_SIZE


class BaseClient(ABC):
    max_timeout = DEFAULT_TIMEOUT_SECONDS
    max_pool_size = DEFAULT_MAX_POOL_SIZE

    @property
    def docker_client(self) -> DockerClient:
        """Docker Client

        Returns:
            DockerClient: docker client object
        """
        return self._docker_client

    @docker_client.setter
    def docker_client(self, client: DockerClient) -> None:
        self._docker_client = client

    @abstractmethod
    def initialise_client(self) -> None:
        """This is the method that initializes the docker client and needs
        be implemented.
        """
        ...



from typing import Any, Dict, List, Optional
from pydantic import BaseModel
from dataclasses import dataclass

from testcompose.models.client import Login
from testcompose.models.volume import VolumeMapping
from testcompose.models.network import ContainerNetworkSettings, NetworkConstants


class ContainerParam(BaseModel):
    image: str
    exposed_ports: List[str]
    command: Optional[str] = None
    entry_point: Optional[str] = None
    environment_variables: Optional[Dict[str, Any]] = None
    volumes: Optional[List[VolumeMapping]] = None
    auto_remove: bool = True
    remove_container: bool = True
    registry_login_param: Login = Login()


class RunningContainer(BaseModel):
    containers: Dict[str, Any] = dict()
    extra_envs: Dict[str, Dict[str, Any]] = dict()
    precedence_map: Dict[str, int] = dict()


class ContainerState(BaseModel):
    Status: str
    Running: bool
    Paused: bool
    Restarting: bool
    OOMKilled: bool
    Dead: bool
    ExitCode: int


class ContainerHostConfig(BaseModel):
    NetworkMode: str = NetworkConstants.DEFAULT_NETWORK_MODE


class RunningContainerAttributes(BaseModel):
    Id: str
    State: ContainerState
    Platform: str
    HostConfig: ContainerHostConfig
    NetworkSettings: ContainerNetworkSettings


@dataclass(frozen=True)
class RunningContainerKeys:
    CONTAINERS: str = "containers"
    EXTRA_ENVS: str = "extra_envs"
    PRECEDENCE_MAP: str = "precedence_map"


@dataclass(frozen=True)
class RunningContainerEnvPrefixes:
    INTERNAL_HOST: str = "DOCKER_PYTHON_INTERNAL_HOST"
    EXTERNAL_HOST: str = "DOCKER_PYTHON_EXTERNAL_HOST"
    MAPPED_PORTS: str = "DOCKER_PYTHON_MAPPED_PORTS"



from typing import Dict, List, Optional
from docker.models.networks import Network
from testcompose.models.network import NetworkConstants
from docker.client import DockerClient


class ContainerNetwork:
    """Network management for running container.This class ensures
    all containers in the same test belong to the same network.
    Contians utility to create and cleanup network.

    Args:
        docker_client (DockerClient): Docker client
        network_name (str): Name of test network
        auto_create_network (bool, optional): Create network if it doesn't exists. Defaults to False.
    """

    def __init__(
        self, docker_client: DockerClient, network_name: str, auto_create_network: bool = False
    ) -> None:
        self._docker_client = docker_client
        self.get_defined_network(network_name, auto_create_network)

    @property
    def container_network(self) -> Network:
        """Network Object

        Returns:
            docker.models.networks.Network: container network object
        """
        return self._container_network

    @container_network.setter
    def container_network(self, network: Network) -> None:
        self._container_network = network

    @property
    def network_name(self) -> Optional[str]:
        """Network Name

        Returns:
            Optional[str]: Network Name
        """
        return self.container_network.name

    def get_defined_network(self, network_name: str, auto_create_network: bool = False):
        """Identify or create container network

        Args:
            network_name (str): Network name
            auto_create_network (bool, optional): If set, create network if it does not exists. Defaults to False.

        Raises:
            AttributeError: When network does not exist or can not be created.
        """
        try:
            networks: List[Network] = self._docker_client.networks.list(names=[network_name])  # type: ignore
        except Exception:
            networks = list()
        if networks:
            self.container_network = networks[0]
        elif not networks and auto_create_network:
            self.container_network = self._create_group_network(network_name)
        else:
            raise AttributeError

    @property
    def container_network_id(self) -> Optional[str]:
        """Container Short Id

        Returns:
            Optional[str]: container_id
        """
        return self.container_network.short_id

    def remove_network(self):
        """Cleanup created network"""
        self.container_network.remove()

    def _create_group_network(self, network_name: str, label: Dict[str, str] = dict()) -> Network:
        return self._docker_client.networks.create(  # type: ignore
            name=network_name,
            driver=NetworkConstants.DEFAULT_NETWORK_MODE,
            check_duplicate=True,
            internal=False,
            labels=label or None,
            enable_ipv6=False,
            attachable=True,
            scope=NetworkConstants.DEFAULT_NETWORK_SCOPE,
        )



from abc import ABC, abstractmethod
import pathlib
from typing import Any, Dict, List, Optional
from docker.client import DockerClient
from testcompose.models.container import ContainerParam
from testcompose.models.volume import VolumeMapping
from docker.errors import ImageNotFound
from testcompose.models.volume import VolumeSourceTypes


class ContainerBuilder(ABC):
    """Container Builder

    Args:
        ABC ([type]):
    """

    def __init__(self, container_param: ContainerParam) -> None:
        self._environments: Dict[str, Any] = dict()
        self._ports: Dict[int, Any] = dict()
        self._volumes: Dict[str, Dict[str, str]] = dict()
        self._docker_host: Optional[str] = None
        self.generic_container_param = container_param

    @property
    def generic_container_param(self) -> ContainerParam:
        """Container parameter

        Returns:
            ContainerParam: parameter model
        """
        return self._container_param

    @generic_container_param.setter
    def generic_container_param(self, param: ContainerParam) -> None:
        self._container_param = param

    @property
    def docker_client(self) -> DockerClient:
        """Docker Client

        Returns:
            DockerClient: client object
        """
        return self._docker_client

    @docker_client.setter
    def docker_client(self, client: DockerClient):
        self._docker_client = client

    def with_exposed_ports(self, ports: Optional[List[str]]):
        """List of exposed port to be assigned random port
        numbers on the host. Random ports are exposed to the
        host. A fixed port can be assigned on the host by providing
        the port in the format **[host_port:container_port]**

        Args:
            ports (Optional[List[str]]): list of container exposed ports
        """
        if ports:
            for port in ports:
                _ports = str(port).split(":")
                if len(_ports) == 2:
                    self._ports[int(_ports[1])] = int(_ports[0])
                else:
                    self._ports[int(port)] = None

    def with_volumes(self, volumes: Optional[List[VolumeMapping]]) -> None:
        """A list of volume mappings to be mounted in the container.

            VolumeMapping:
                host: host volume path or a docker volume name
                container: path to mount the host volume in the container
                mode: volume mode [ro|rw]
                source: source of the volume [local|docker]
                        local: a file or directory to be mounted
                        docker: a docker volume (existing or to be created)

        Args:
            volumes (Optional[List[VolumeMapping]]): Optional list of volumes to mount on the container
        """
        if volumes:
            for vol in volumes:
                host_bind: Optional[str] = None
                if vol.source == VolumeSourceTypes.DOCKER_SOURCE:
                    host_bind = vol.host
                elif vol.source == VolumeSourceTypes.LOCAL_SOURCE:
                    host_bind = str(pathlib.Path(vol.host).absolute())
                if not host_bind:
                    raise ValueError("Volume source can only be one of local|docker")
                self._volumes[host_bind] = {"bind": vol.container, "mode": vol.mode}

    def with_environment(self, env: Optional[Dict[str, Any]]):
        """Environment variables for running containers

        Args:
            env (Optional[Dict[str, Any]]): [description]
        """
        if env:
            for k, v in env.items():
                self._environments[k] = v

    def pull_image(self, image_pull_policy="Always_Pull"):
        """Pull image of registry
        Pull polic can be one of the following:

            image_pull_policy:
                - Always_Pull [default]
                - Never

        Args:
            image_pull_policy (str, optional): Image pull policy for the test.
        """
        # TODO: Allow pull toggle
        try:
            self.docker_client.images.get(name=self.generic_container_param.image)
        except ImageNotFound:
            self.docker_client.images.pull(repository=self.generic_container_param.image)

    def build(self, docker_client: DockerClient) -> None:
        """Container parameters

        Args:
            docker_client (DockerClient): docker client
        """
        self.docker_client = docker_client
        self.with_environment(self._container_param.environment_variables)
        self.with_volumes(self._container_param.volumes)
        self.with_exposed_ports(self._container_param.exposed_ports)
        if self.generic_container_param.registry_login_param.username:
            login = self.docker_client.login(**self.generic_container_param.registry_login_param.dict())
            print(login)
        print(f"Pulling image {self.generic_container_param.image}")
        self.pull_image()

    @abstractmethod
    def start(self):
        ...

    @abstractmethod
    def stop(self):
        ...



from typing import Dict, List, Set, Tuple
from testcompose.models.config import ITestConfig, ITestConfigServices, RankedServices, RankedServiceKey
from testcompose.log_setup import stream_logger

logger = stream_logger(__name__)


class Config:
    """This class consumes the model created from a config file.
    This is an important class that sets the precedence of how the
    different containers are to be started and stopped. Usually, the
    precedence are set correctly if the `depends_on` parameter of the
    config is set. Cyclic dependency will fail the test before it starts.

    Args:
        test_services (ITestConfig): model resulting from a parsed configuration file.
    """

    def __init__(self, test_services: ITestConfig) -> None:
        self._get_container_spawn_precedence(test_services)

    @property
    def ranked_itest_config_services(self) -> RankedServices:
        """Object containing the ordered services from the config

        Returns:
            RankedServices: ranked container services
        """
        return self._ranked_it_services

    @ranked_itest_config_services.setter
    def ranked_itest_config_services(self, ranked_services: RankedServices) -> None:
        self._ranked_it_services = ranked_services

    def _get_container_spawn_precedence(self, test_services: ITestConfig) -> None:
        """

        Args:
            test_services (ITestConfig): model resulting from a parsed configuration file.

        Raises:
            ValueError: raised if test_services is `null`
            AttributeError: raised if no concreate networking is provided
        """
        _config_services: Dict[Tuple[int, str], ITestConfigServices] = dict()
        if not test_services:
            logger.error("Config content can not be Null")
            raise ValueError

        if not test_services.services:
            logger.error("No service was found in the provided config")
            raise ValueError

        _processed_containers: List[ITestConfigServices] = self._compute_container_ranks(
            processed_containers=list(),
            processed_container_names=set(),
            unprocessed_containers=test_services.services,
        )

        for _rank, _service in enumerate(_processed_containers):
            _config_services.update({(_rank, _service.name): _service})

        if all([test_services.network.auto_create, test_services.network.use_random_network]):
            raise AttributeError(
                "Both auto_create and use_random_network options of network can not be True!!"
            )

        self.ranked_itest_config_services = RankedServices(
            **{RankedServiceKey.SERVICES: _config_services, RankedServiceKey.NETWORK: test_services.network}
        )

    def _compute_container_ranks(
        self,
        processed_containers: List[ITestConfigServices],
        processed_container_names: Set[str],
        unprocessed_containers: List[ITestConfigServices],
    ) -> List[ITestConfigServices]:
        """The main method that computes the ranking of the services specified
        in the config.

        Args:
            processed_containers (List[ITestConfigServices]): Container object of the services to be tested
            processed_container_names (Set[str]): the names of the services as specified in the configs that had been ranked
            unprocessed_containers (List[ITestConfigServices]): the names of the services as specified in
                                                                the configs yet to be ranked

        Raises:
            AttributeError: raised to prevent both unprocessed_containers and processed_containers from being `null`
            ValueError: rasied when cyclic dependency is detected

        Returns:
            List[ITestConfigServices]: A list of ranked service models.
        """
        if not unprocessed_containers and not processed_containers:
            raise AttributeError("Processed and Unprocessed container are both empty")

        if not unprocessed_containers:
            return processed_containers
        else:
            passes: Set[str] = set()
            for _container in unprocessed_containers:
                passes.add(_container.name)
                if not _container.depends_on:
                    processed_containers.append(_container)
                    processed_container_names.add(_container.name)
                else:
                    if set(_container.depends_on).issubset(processed_container_names):
                        processed_containers.append(_container)
                        processed_container_names.add(_container.name)
                    elif not set(_container.depends_on).issubset(processed_container_names) and set(
                        _container.depends_on
                    ).issubset(passes):
                        raise ValueError(
                            f"Cyclic container dependencies detected: {processed_container_names} <=> {set(_container.depends_on)}"
                        )
            passes.clear()
            del passes
            return self._compute_container_ranks(
                processed_containers,
                processed_container_names,
                [x for x in unprocessed_containers if x.name not in processed_container_names],
            )



from typing import Dict, List, Optional
from dataclasses import dataclass
from pydantic import BaseModel


class HostPortMapping(BaseModel):
    HostIp: str
    HostPort: str


class ContainerNetworks(BaseModel):
    Aliases: Optional[List[str]] = None
    NetworkID: str
    EndpointID: str
    Gateway: str
    IPAddress: str


class ContainerNetworkSettings(BaseModel):
    Ports: Dict[str, Optional[List[HostPortMapping]]] = dict()
    Networks: Dict[str, ContainerNetworks]


@dataclass(frozen=True)
class NetworkConstants:
    DEFAULT_NETWORK_MODE: str = "bridge"
    DEFAULT_NETWORK_SCOPE: str = "local"



import re
import socket
from docker.client import DockerClient
from requests import get
from datetime import datetime
from time import sleep
from typing import Any, ByteString, Dict, List, Match, Optional, Tuple, Union
from docker.models.containers import Container
from docker.errors import APIError
from testcompose.models.network import HostPortMapping, NetworkConstants
from testcompose.containers.container_builder import ContainerBuilder
from testcompose.models.container import ContainerParam, RunningContainerAttributes


class GenericContainer(ContainerBuilder):
    """Generic container object to control the base properties of
    running containers. This class manipulates the base components
    of a container.

    Parameters not part of the **ContainerParam** object can be provided as
    keyword arguments.


    Args:
        docker_client (DockerClient): docker client
        container_param (ContainerParam): container parameter for a given service
    """

    def __init__(self, docker_client: DockerClient, container_param: ContainerParam, **kwargs) -> None:
        super().__init__(container_param)
        self._kwargs = kwargs
        self.env = container_param.environment_variables or dict()
        self.build(docker_client=docker_client)

    @property
    def env(self) -> Dict[str, Any]:
        """Container environment variables

        Returns:
            Dict[str, Any]: environment variables
        """
        return self._env

    @env.setter
    def env(self, env_variable: Dict[str, Any]):
        self._env = env_variable

    @property
    def container(self) -> Container:
        return self._container

    @container.setter
    def container(self, _container) -> None:
        self._container = _container

    @property
    def get_container_attr(self) -> RunningContainerAttributes:
        """Running container attributes

        Returns:
            RunningContainerAttributes: Container attribute object
        """
        return self._container_attr

    @get_container_attr.setter
    def get_container_attr(self, atrr: RunningContainerAttributes):
        """Running container attributes. Execute reload() to refresh this
        property.

        Args:
            atrr (RunningContainerAttributes): container attributes
        """
        self._container_attr = atrr

    def start(self) -> None:
        """Start a container"""
        try:
            if self.docker_client:
                self.container = self.docker_client.containers.run(
                    image=self.generic_container_param.image,
                    command=self.generic_container_param.command,
                    detach=True,
                    environment=self._environments,
                    ports=self._ports,
                    volumes=self._volumes,
                    entrypoint=self.generic_container_param.entry_point,
                    auto_remove=self.generic_container_param.auto_remove,
                    remove=self.generic_container_param.remove_container,
                    **self._kwargs,
                )
                self.reload()
        except Exception as exc:
            print(exc)
            self.stop()

    def stop(self, force=True, delete_volume=True) -> None:
        """Stop a running container

        Args:
            force (bool, optional): [description]. Defaults to True.
            delete_volume (bool, optional): [description]. Defaults to True.
        """
        try:
            if self.container:
                self.container.remove(v=delete_volume, force=force)
        except APIError as exc:
            print(exc)

    def reload(self) -> None:
        """Reload the attributes of a running container"""
        self.container.reload()
        self.get_container_attr = RunningContainerAttributes(**self.container.attrs)
        self.wait_on_condition()

    def get_exposed_port(self, port: str) -> str:
        """Get host port bound to the container exposed port

        Args:
            port (str): container exposed port

        Returns:
            str: Host port bound to the container exposed port
        """
        return self.get_mapped_container_ports([port])[port]

    def get_mapped_container_ports(self, ports: List[str]) -> Dict[str, str]:
        """Host port bound to the container returns as a k/v of the
        container exposed port as key and the host bound port as the value.

        Args:
            ports (List[str]): List of container exposed port to be mapped to host port

        Returns:
            Dict[str, str]: Mapped container-host ports.
        """
        mapped_ports: Dict[str, str] = dict()
        _ports: Dict[str, Optional[List[HostPortMapping]]] = self.get_container_attr.NetworkSettings.Ports
        for cport, hport in _ports.items():
            host_port = re.sub(f"[^0-9]", "", cport)
            if host_port in ports and hport:
                mapped_ports.update({host_port: str(hport[0].HostPort)})
        return mapped_ports

    def get_container_id(self) -> Optional[str]:
        """Container id

        Returns:
            Optional[str]: Container Id
        """
        if self.container:
            return self.container.id
        return None

    def get_container_ip(self, network_name=NetworkConstants.DEFAULT_NETWORK_MODE) -> str:
        """Running container IP address

        Args:
            network_name ([type], optional): [description]. Defaults to NetworkConstants.DEFAULT_NETWORK_MODE.

        Raises:
            ValueError: when network_name is not provided

        Returns:
            str: container IP Address
        """
        if not network_name:
            raise ValueError("Docker network name must be provided!!")
        _networks = self.get_container_attr.NetworkSettings.Networks[network_name]

        return _networks.IPAddress

    def exe_command(self, command: Union[str, List[str]]) -> Tuple[int, ByteString]:
        """Execute a command inside a container once it started running.

        Args:
            command (Union[str, List[str]]): command to run in the container

        Raises:
            RuntimeError: when the container object is not set

        Returns:
            Tuple[int, ByteString]: A tuple of (exit_code, output)
        """
        if not self.container:
            raise RuntimeError("Container must already be running to exec a command")
        return self.container.exec_run(cmd=command)

    def search_container_logs(self, search_string: str, timeout: float = 300.0, interval: int = 1) -> bool:
        """Search for a given predicate in the container log. Useful to check if a
        container is running and healthy

        Args:
            search_string (str): predicate to search in the log
            timeout (float, optional): Defaults to 300.0.
            interval (int, optional): Defaults to 1.

        Raises:
            ValueError: if a non string predicate is passed

        Returns:
            bool: True if the log contains the provided predicate
        """
        if not isinstance(search_string, str):
            raise ValueError

        prog = re.compile(search_string, re.MULTILINE).search
        start = datetime.now()
        output: Optional[Match[str]] = None
        while (datetime.now() - start).total_seconds() < timeout:
            output = prog(self.container.logs().decode())
            if output:
                return True
            if (datetime.now() - start).total_seconds() > timeout:
                return False
            sleep(interval)
        print("%s", self.container.logs().decode())
        return True if output else False

    @staticmethod
    def get_container_host_ip() -> str:
        """The host IP where the container runs

        Returns:
            str: host IP
        """
        return socket.gethostbyname(socket.gethostname())

    def with_wait_for_http(
        self, http_port: str, status_code: int = 200, end_point: str = "/", server_startup_time: int = 20
    ) -> bool:
        """Endpoint health-check for a container. A running service
        with an exposed endpoint is queried and the response code is
        checked with the expected response code.

        Args:
            http_port (str): container service port
            status_code (int, optional): Defaults to 200.
            end_point (str, optional): Provided service endpoint. Defaults to "/".
            server_startup_time (int, optional): Expected wait time for the service to start. Defaults to 20.

        Returns:
            bool: Endpoint returned expected status code
        """
        response_check: bool = True
        for _ in range(0, 3):
            sleep(server_startup_time)
            try:
                site_url = f"http://{self.get_container_host_ip()}:{self.get_exposed_port(port=http_port)}/{end_point.lstrip('/')}"
                response = get(url=site_url.rstrip("/"))
                if response.status_code != status_code:
                    print("%s", self.container.logs().decode())
                    response_check = False
                break
            except Exception as exc:
                response_check = False
                print("HTTP_CHECK_ERROE: %s", exc)
        return response_check

    def wait_on_condition(self, status="running", delay=0.1, timeout=40):
        """Method useful for checking a running container status to
        allow for fetching the latest attribute from the container

        Args:
            status (str, optional): Status to check for. Defaults to "running".
            delay (float, optional): Delay time before the next check. Defaults to 0.1.
            timeout (int, optional): Defaults to 40.
        """

        def condition() -> bool:
            """Container emits expected status

            Returns:
                bool: container matches expected status
            """
            return status == self.get_container_attr.State.Status

        start = datetime.now()
        while not condition():
            if (datetime.now() - start).total_seconds() > timeout:
                raise AssertionError("Container status %s not obtained after 40secx" % status)
            sleep(delay)
        print(f"Found Status {self.get_container_attr.State.Status}")
        return



from pydantic import BaseModel
from dataclasses import dataclass


class VolumeMapping(BaseModel):
    host: str
    container: str
    mode: str
    source: str = "docker"


@dataclass(frozen=True)
class VolumeSourceTypes:
    LOCAL_SOURCE: str = "local"
    DOCKER_SOURCE: str = "docker"



from docker.client import DockerClient
from .base_client import BaseClient
from testcompose.models.client import ClientFromUrl


class UrlClient(BaseClient):
    def __init__(self, client_param: ClientFromUrl) -> None:
        """Client url parameters

        Args:
            client_param (ClientFromUrl): model class consisting of parameters for client connections with http url
        """
        super().__init__()
        self._docker_host = client_param.docker_host
        self._version = client_param.version
        self._timeout = client_param.timeout or BaseClient.max_timeout
        self._tls = client_param.tls
        self._user_agent = client_param.user_agent
        self._credstor_env = client_param.credstor_env
        self._use_ssh_client = client_param.use_ssh_client
        self._max_pool_size = client_param.max_pool_size or BaseClient.max_pool_size

        # self.docker_client = self.initialise_client()

    def initialise_client(self) -> None:
        """Init docker client

        Returns:
            DockerClient: docker client
        """
        self.docker_client = DockerClient(
            base_url=self._docker_host,
            version=self._version,
            timeout=self._timeout,
            tls=self._tls,
            user_agent=self._user_agent,
            credstor_env=self._credstor_env,
            use_ssh_client=self._use_ssh_client,
            max_pool_size=self._max_pool_size,
        )



from dataclasses import dataclass
from pydantic import BaseModel
from typing import Any, Dict, List, Optional, Tuple
from .volume import VolumeMapping
from .network import NetworkConstants


class LogWaitParameters(BaseModel):
    log_line_regex: str
    wait_timeout: str
    poll_interval: str


class HttpWaitParameters(BaseModel):
    http_port: int
    response_status_code: int
    startup_delay_time: int = 20
    end_point: str = "/"


class ITestConfigServices(BaseModel):
    name: str
    image: str
    exposed_ports: List[str]
    labels: List[str] = list()
    volumes: List[VolumeMapping] = list()
    depends_on: List[str] = list()
    environment: Dict[str, Any] = dict()
    auto_remove: bool = True
    remove_container: bool = True
    log_wait_parameters: Optional[LogWaitParameters] = None
    http_wait_parameters: Optional[HttpWaitParameters] = None
    command: Optional[str] = None
    entrypoint: Optional[str] = None


class ConfigCustomNetwork(BaseModel):
    name: str = NetworkConstants.DEFAULT_NETWORK_MODE
    auto_create: bool = False
    use_random_network: bool = False


class ITestConfig(BaseModel):
    services: List[ITestConfigServices]
    network: ConfigCustomNetwork = ConfigCustomNetwork()


class RankedServices(BaseModel):
    services: Dict[Tuple[int, str], ITestConfigServices]
    network: ConfigCustomNetwork


@dataclass(frozen=True)
class RankedServiceKey:
    SERVICES: str = "services"
    NETWORK: str = "network"



import docker
from .base_client import BaseClient
from testcompose.models.client import ClientFromEnv


class EnvClient(BaseClient):
    def __init__(self, client_param: ClientFromEnv) -> None:
        """Client environment parameters

        Args:
            client_param (ClientFromEnv): model class consisting of parameters for client connections with environment variables
        """
        super(EnvClient, self).__init__()
        self._version = client_param.version
        self._timeout = client_param.timeout or BaseClient.max_timeout
        self._max_pool_size = client_param.max_pool_size or BaseClient.max_pool_size
        self._use_ssh_client = client_param.use_ssh_client
        self._ssl_version = client_param.ssl_version
        self._assert_hostname = client_param.assert_hostname
        self._environment = client_param.environment

        # self.docker_client = self.initialise_client()

    def initialise_client(self) -> None:
        """Init docker client

        Returns:
            DockerClient: docker client
        """
        self.docker_client = docker.from_env(
            version=self._version,
            timeout=self._timeout,
            max_pool_size=self._max_pool_size,
            use_ssh_client=self._use_ssh_client,
            ssl_version=self._ssl_version,
            assert_hostname=self._assert_hostname,
            environment=self._environment,
        )



import logging


def stream_logger(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)
    return logger



from typing import Any, Dict, Optional
from pydantic import BaseModel


class ClientFromEnv(BaseModel):
    version: str = "auto"
    timeout: Optional[int] = None
    max_pool_size: Optional[int] = None
    use_ssh_client: bool = False
    ssl_version: Optional[int] = None
    assert_hostname: Optional[bool] = None
    environment: Optional[Dict[str, Any]] = None


class ClientFromUrl(BaseModel):
    docker_host: Optional[str] = None
    version: str = "auto"
    timeout: Optional[int] = None
    tls: Optional[bool] = None
    user_agent: Optional[str] = None
    credstor_env: Optional[Dict[str, Any]] = None
    use_ssh_client: Optional[bool] = None
    max_pool_size: Optional[int] = None


class Login(BaseModel):
    username: Optional[str] = None
    password: Optional[str] = None
    email: Optional[str] = None
    registry: Optional[str] = None
    reauth: Optional[bool] = False
    dockercfg_path: Optional[str] = None
